import direct.directbase.DirectStart
from direct.gui.OnscreenText import OnscreenText
from direct.gui.DirectGui import *
from pandac.PandaModules import *
from direct.task import Task
from direct.directtools.DirectGeometry import LineNodePath
from direct.showbase.DirectObject import *
from direct.interval.IntervalGlobal import *

import psyco
from psyco.classes import __metaclass__
import sys, os, random, math, time, string

## GLOBALS
PATH = os.path.abspath(sys.path[0]) 
PATH = Filename.fromOsSpecific(PATH).getFullpath() 
DATA_PATH = os.path.join(PATH, "data")
UPDATE_FREQ = 20  # times a second
UPDATE_TIME = 1.0 / UPDATE_FREQ
CONFIG_FILE = "textinput.txt"
global frameCount
frameCount = 0
global parSeq
parSeq = Parallel()  # parallel sequences


# boids control
NUM_BOIDS = 100
SEPARATION = 2   #minimum separation
global SPEED_LIMIT
SPEED_LIMIT = 2
CORNER_DISTANCE_SAME = 10
CORNER_DISTANCE_DIFF = 30
CORNER_STRENGTH = 1 #the lower the number, the higher the strength
MAGNET_DISTANCE = 20
MAGNET_STRENGTH = 20  #the lower the number, the higher the strength
Y_POS = -0.5  #how much above the paper everything is
WIND_STRENGTH = 50 #lower=stronger

# bounds for boids
global X_MAX, X_MIN, Z_MAX, Z_MIN

BINS_X = 10
BINS_Z = 10

# signifies the corners of the paper
BOTTOM_LEFT  = 0 
BOTTOM_RIGHT = 2
TOP_RIGHT    = 1
TOP_LEFT     = 3


# make a pool of all the letter models so that
# they dont have to be reloaded more than once
ModelPool.loadModel(DATA_PATH+"/a.egg")
ModelPool.loadModel(DATA_PATH+"/b.egg")
ModelPool.loadModel(DATA_PATH+"/c.egg")
ModelPool.loadModel(DATA_PATH+"/d.egg")
ACCEPTED_LETTERS = ['a','b','c','d']


# used for quickly determining neighboring boids
# based on bin-lattice space partitioning method
class Bins:
    def __init__(self, paper):
        # first create a three dimensional array - hack it with a hash
        self.paper = paper
        self.__bins = {}  
        for x in range(BINS_X):
            for z in range(BINS_Z):
                self.__bins[x,z] = {}
                
        # size of each bin
        self.binW = (X_MAX - X_MIN)/BINS_X
        self.binH = (Z_MAX - Z_MIN)/BINS_Z
        # bounds
        self.minX = X_MIN
        self.maxX = X_MAX
        self.minZ = Z_MIN
        self.maxZ = Z_MAX
        # number of bins
        self.numX = BINS_X
        self.numZ = BINS_Z

        self.doDrawBins = False
        self.__drawBins()
        

        # mostly for debugging purposes
    def showBins(self):
        self.doDrawBins = not self.doDrawBins
        if self.doDrawBins:
            self.lines.show()
        else:
            self.lines.hide()

    def __drawBins(self):
        y = Y_POS
        pointList = [[[self.minX,y,self.minZ], [self.maxX,y,self.minZ]],
                     [[self.minX,y,self.minZ], [self.minX,y,self.maxZ]],
                     [[self.maxX,y,self.minZ], [self.maxX,y,self.maxZ]],
                     [[self.minX,y,self.maxZ], [self.maxX,y,self.maxZ]]]

        # vertical lines
        for x in range(self.minX, self.maxX, self.binW):
            pointList.append([ [x,y,self.minZ], [x,y,self.maxZ] ])
        
        # horizontal lines
        for z in range(self.minZ, self.maxZ, self.binH):
            pointList.append([ [self.minX,y,z], [self.maxX,y,z] ])
                       
        self.lines = LineNodePath()
        self.lines.reparentTo(self.paper)
        self.lines.setColor(0,0,0)
        self.lines.setThickness(2)
        self.lines.drawLines(pointList)
        self.lines.create()
        self.lines.hide()


        # adds the boid for the first time
    def addBoid(self, boid):
        bin = self.__getIndex(boid.pos.getX(), boid.pos.getY())
        self.__bins[bin[0],bin[1]][boid] = 1
        boid.bin = bin
        

        # removes the boid from the bins
    def removeBoid(self, boid):
        bin = self.__getIndex(boid.pos.getX(), boid.pos.getY())
        del self.__bins[bin[0],bin[1]][boid]

        
        # puts the boid in the right bin based on its position
    def updateBoid(self, boid):
        newBin = self.__getIndex(boid.pos.getX(), boid.pos.getY())
        oldBin = boid.bin

        if oldBin == newBin:     # no change, most common occurence
            pass                       
        else:                    # boid moved from one bin to another
            del self.__bins[oldBin[0],oldBin[1]][boid]
            self.__bins[newBin[0],newBin[1]][boid] = 1
            boid.bin = newBin


        # returns an index into self.__bins based on the coordinates
        # or None if not in any
    def __getIndex(self, x, z):
        dx = x-self.minX
        dz = z-self.minZ
        i = int( math.floor(dx/self.binW) )
        j = int( math.floor(dz/self.binH) )

        # check whether boid went out of bounds, if it did just threshold it to 0 or max
        if i<0: i=0
        elif i>=self.numX: i=self.numX-1
        if j<0: j=0
        elif j>=self.numZ: j=self.numZ-1

        return (i,j)


        # returns a list of neighboring boids
        # basically all the boids in the neighboring bins
    def getNeighbors(self, boid):
        l = []
        (i,j) = boid.bin 

        # determine the neighboring cells
        indexes = [(i-1,j+1), (i,j+1), (i+1,j+1),
                   (i-1,j),   (i,j),   (i+1,j),
                   (i-1,j-1), (i,j-1), (i+1,j-1)]

        # aggregate all the boids from the neighboring cells into one list
        for ind in indexes:
            try:
                l.extend(self.__bins[ind[0],ind[1]].keys())  # create a list of all the neighbors
            except:
                continue
        
        return l


        
        
class BoidList:
    def __init__(self, paper, letters, bins, magnets, corners, handlerQ, world):
        self.paper = paper
        self.world = world
        self.doPause = False
        self.handlerQ = handlerQ
        self.__letters = letters
        self.__bins = bins
        self.__boids = {} # keys = letters, values = list of boids of that letter
        self.__createBoids()
        self.__lastUpdateTime = UPDATE_TIME
        self.__magnets = magnets
        self.__corners = corners
        self.__toDelete = {}  # boids scheduled for deletion
        self.collisions = {}
        self.inCollisions = {}
         

    def __createBoids(self):
        for letter, numLetters in self.__letters.iteritems():
            l = []
            for i in range(0, numLetters):
                b = Boid(self.paper, letter, self.handlerQ)
                l.append(b)
                self.__bins.addBoid(b)
            self.__boids[letter] = l


        # pause the boids only
    def pause(self):
        self.doPause = not self.doPause


        # loop through all the boids and update their positions based on flocking rules
    def updateBoids(self, task):
        if (task.time - self.__lastUpdateTime) < UPDATE_TIME or self.doPause:
            return Task.cont
        else:
            self.__lastUpdateTime = task.time
            global frameCount
        frameCount += 1

        # check for collisions and store them in a temp hash
        
        #self.checkCollisions()

        # traverse all the boids and update their positions
        # based on various influences in the world
        neighbors = []
        nb = 0
        wind = self.__wind()
        for letter in self.__letters.iterkeys():
            for boid in self.__boids[letter]:   #temp
                neighbors = self.__bins.getNeighbors(boid)
                nb +=1
                # apply all the rules
                oldVel = boid.vel
                boid.vel += self.__cohesion(boid, neighbors)
                boid.vel += self.__separation(boid, neighbors)
                boid.vel += self.__alignment(boid, neighbors)
                boid.vel += wind
                boid.vel += self.__checkMagnet(boid)
                v = self.__checkCorner(boid)
                if not v:   # if boid is in the corner, kill it
                    self.__killBoid(boid)
                    continue
                else:
                    boid.vel += v
                boid.vel += self.__enforceBounds(boid)
                if boid.node in self.collisions:
                    if not self.collisions[boid.node]:
                        boid.vel += self.__turnAround(boid)#self.__moveAwayFromPlace(self.collisions[boid.node], boid)
                        self.collisions[boid.node] = True
                    else:  # we are still inside an object
                        boid.vel = oldVel
                boid.limitSpeed(SPEED_LIMIT)

                # apply the changes (if any)
                boid.setPos(boid.pos + boid.vel)
                self.__bins.updateBoid(boid)  #update its bin memebership

        # update the dying boids
        for b in self.__toDelete.keys():
            if not b.i:
                b.die()  # start the die animation
            elif b.i.isStopped():  # done with animation, kill it
                b.node.removeNode()
                del self.__toDelete[b]
                del b

        if nb==0 and len(self.__toDelete)==0:
            self.world.stopTimer()
            return Task.done
        
        return Task.cont


    def handleInCollision(self, cEntry):
        nodePath = cEntry.getIntoNodePath().findNetTag('boid')
        if not nodePath.isEmpty():
            #print "collided with letter: ", nodePath.getTag('boid')
            pos = cEntry.getSurfacePoint(self.paper)
            self.collisions[nodePath] = False #Vec2D(pos[0], pos[2])
            

    def handleOutCollision(self, cEntry):
        nodePath = cEntry.getIntoNodePath().findNetTag('boid')
        if not nodePath.isEmpty():
            #print "collided with letter: ", nodePath.getTag('boid')
            #pos = cEntry.getSurfacePoint(self.paper)
            if nodePath in self.collisions:
                del self.collisions[nodePath]
            

    def checkCollisions(self):
        base.cTrav.traverse(self.paper)
        numCollisions = self.handlerQ.getNumEntries()
        for i in range(0, numCollisions):
            cEntry = self.handlerQ.getEntry(i)
            nodePath = cEntry.getIntoNodePath().findNetTag('boid')
            if not nodePath.isEmpty():
                pos = cEntry.getSurfacePoint(self.paper)
                self.collisions[nodePath] = Vec2D(pos[0], pos[2])
                #print "collision detected with letter: ", nodePath.getTag('boid')
            

    # rule 1: keep the boids close together
    def __cohesion(self, boid, neighbors):
        pc = Vec2D(0,0)
        for b in neighbors:
            if boid != b:
                pc += b.pos

        pc = pc/len(neighbors)
        return (pc - boid.pos) / 50


    # rule 2: keep the boids a small distance apart
    def __separation(self, boid, neighbors):
        c = Vec2D(0,0)
        for b in neighbors:
            if boid != b:
                diffPos = boid.pos - b.pos
                if diffPos.length() < SEPARATION:
                    c = c + diffPos
        return c


    # rule 3: keep the boids moving at similar speeds
    def __alignment(self, boid, neighbors):
        pv = Vec2D(0,0)
        for b in neighbors:
            if boid != b:
                pv += b.vel

        pv = pv/len(neighbors)
        return (pv - boid.vel) / 8


    # rule 4: enforce bounds on boids
    def __enforceBounds(self, boid):
        v = Vec2D(0,0)
	diff = 0.5  #direction adjustment magnitude
	
        if boid.pos.getX() < X_MIN:
            v.setX(diff)
	elif boid.pos.getX() > X_MAX:
            v.setX(-diff)

        if boid.pos.getY() < Z_MIN:
            v.setY(diff)
	elif boid.pos.getY() > Z_MAX:
            v.setY(-diff)

##         if boid.pos.getZ() < Z_MIN:
##             v.setZ(diff)
## 	elif boid.pos.getZ() > Z_MAX:
##             v.setZ(-diff)

	return v


    # rule 5: move towards a point
    def __moveToPlace(self, place, boid):
        return (place - boid.pos) / 1


    # rule 6: move away from a point
    def __moveAwayFromPlace(self, place, boid):
        return self.__moveToPlace(place, boid)*(-1)


    # turns the boid in the opposite direction
    def __turnAround(self, boid):
        #frame = globalClock.getFrameCount()
        #if frame - boid.lastCol < 95:
        #    return Vec2D(0,0)
        #else:
        #print "turning boid around"
        #boid.lastCol = frame
        return boid.vel*(-5)


    def __wind(self):
        hpr = self.paper.getHpr()/WIND_STRENGTH
        return Vec2D(hpr[0], hpr[1]*(-1))
        

    # move towards the magnet if it's close enough
    def __checkMagnet(self, boid):
        for l, m in self.__magnets.iteritems():
            if not m.held:
                continue
            mp = m.pos
            bp = boid.pos
            if (bp - mp).length() < MAGNET_DISTANCE:
                if boid.letter == l:
                    return (mp-bp)/MAGNET_STRENGTH
                else:
                    return (mp-bp)/(MAGNET_STRENGTH*(-3))
            else:
                continue
        return Vec2D(0,0)

    # boids get attracted by the corners
    # if they get in it, they die
    def __checkCorner(self, boid):
        for l, c in self.__corners.iteritems():
            cp = c.pos
            bp = boid.pos
            dist = (bp - cp).length()
            if boid.letter == l and dist < 2:
                return False
            elif dist < CORNER_DISTANCE_SAME:
                if boid.letter == l:# and dist<(CORNER_DISTANCE-5):
                    return (cp-bp)/CORNER_STRENGTH
                else:
                    return Vec2D(0,0)
            elif dist < CORNER_DISTANCE_DIFF:
                if boid.letter == l:
                    return Vec2D(0,0)
                else:
                    return (cp-bp)/(CORNER_STRENGTH*(-1))
            else:
                continue
        return Vec2D(0,0)

    # removes the boid from the lists and schedule it for deletion
    def __killBoid(self, boid):
        self.__bins.removeBoid(boid)
        self.__boids[boid.letter].remove(boid)
        self.__toDelete[boid] = 1



    
# a class that describes boids
class Boid:
    def __init__(self, paper, letter, handlerQ):
        self.letter = letter
        self.handlerQ = handlerQ
        self.paper = paper #parent
        self.bin = None   # a tuple containing the index into bins table
        self.i = None # None meaning it hasnt even started
        self.__loadModel()
        self.lastCol = 0  #frame count of the last collision
        self.collided = False
        
        # initialize position and velocity
        random.seed()
        self.setPos( Vec2D(random.randint(-20, 20), random.randint(-20, 20)) )
        self.vel = Vec2D(0,0)
            
    def __loadModel(self):
        self.node = loader.loadModelCopy(DATA_PATH+"/"+self.letter+".egg")
        self.node.reparentTo(self.paper)
        bounds = self.node.getBounds()
        self.node.setScale(0.3, 0.3, 0.3)
        self.node.setAllColorScale(0.3)
        self.node.setTag('boid', self.letter)
        
        # collisions
        cs = CollisionSphere(0,0,0, bounds.getRadius())
        cnode = CollisionNode('sphere_boid')
        cnode.addSolid(cs)
        cnode.setIntoCollideMask(BitMask32(0x10))
        cnodePath = self.node.attachNewNode(cnode)
        #cnodePath.show()

    def setPos(self, vec):
        self.pos = vec
        self.node.setFluidPos(vec[0], Y_POS, vec[1])

    def setVel(self, vec):
        self.vel = vec

    def getVel(self):
        return self.vel

    def getPos(self):
        return self.pos

    def limitSpeed(self, limit):
        if self.vel.length() > limit:
            self.vel = (self.vel / self.vel.length()) * limit

    def die(self):
        self.i = LerpScaleInterval(self.node, 2, Vec3(0.1,0.1,0.1))
        self.i.start()



class Magnet:
    def __init__(self, letter, cornerObj, paper):
        self.letter = letter
        self.corner = cornerObj
        self.paper = paper
        self.pos = self.corner.pos  #Vec2D
        self.pPos = self.pos
        self.loadModel()
        self.setupCollisions()
        self.held = False

    def loadModel(self):
        self.node = loader.loadModelCopy(DATA_PATH+"/"+self.letter+"_magnet.egg")#"+letter+".egg")
        self.node.reparentTo(self.paper)
        #self.node.setHpr(0, 90,0)
        self.node.setAllColorScale(0.3) #setColorScale(0.2)#0.6, 0.6, 0.6, 0)
        self.node.setPos(self.pos[0], Y_POS-1, self.pos[1])
        self.node.setTag("magnet", self.letter)
        
    def setupCollisions(self):
        bounds = self.node.getTightBounds()
        x = (bounds[1].getX() - bounds[0].getX()) / 2
        z = (bounds[1].getZ() - bounds[0].getZ()) / 2
        cs = CollisionSphere(0,0,0, max(x,z))
        cnode = CollisionNode('sphere')
        cnode.addSolid(cs)
        cnode.setIntoCollideMask(BitMask32(0x2))
        cnodePath = self.node.attachNewNode(cnode)
        #cnodePath.show()

    def setPosition(self, pos):
        self.pos = pos
        self.node.setFluidPos(pos[0], Y_POS-1, pos[1])
        #self.node.lookAt(0,0,0)#self.pPos[0], Y_POS, self.pPos[1])
        #self.node.setHpr(0,0,90)
        self.pPos = pos

    def moveToCorner(self):
        self.pos = self.corner.pos
        self.node.setPos(self.pos[0], Y_POS-1, self.pos[1])
        self.pPos = self.pos



class Corner:
    def __init__(self, corner, letter, paper):
        self.letter = letter
        self.corner = corner
        self.paper = paper
        self.__setPosition()
        self.loadModel()

    def loadModel(self):
        self.node = loader.loadModel(DATA_PATH+"/"+self.letter+"_corner.egg")
        self.node.reparentTo(self.paper)
        bounds = self.node.getTightBounds()
        x = (bounds[1].getX() - bounds[0].getX()) / 2
        z = (bounds[1].getZ() - bounds[0].getZ()) / 2
        s = 0.7
        self.node.setScale(s,s,s)
        self.node.setAllColorScale(0.3)

        if self.corner == BOTTOM_LEFT:
            pos = self.pos + Vec2D(x, z)
        elif self.corner == BOTTOM_RIGHT:
            pos = self.pos + Vec2D(-x, z)
        elif self.corner == TOP_RIGHT:
            pos = self.pos + Vec2D(-x, -z)
        elif self.corner == TOP_LEFT:
            pos = self.pos + Vec2D(x, -z)
        
        self.node.setPos(pos[0], Y_POS, pos[1])

    def __setPosition(self):
        if self.corner == BOTTOM_LEFT:
            self.pos = Vec2D(X_MIN, Z_MIN)
        elif self.corner == BOTTOM_RIGHT:
            self.pos = Vec2D(X_MAX, Z_MIN)
        elif self.corner == TOP_RIGHT:
            self.pos = Vec2D(X_MAX, Z_MAX)
        elif self.corner == TOP_LEFT:
            self.pos = Vec2D(X_MIN, Z_MAX)


class Line:
    def __init__(self, start, end, paper, handlerQ):
        self.pointList = [[[start.getX(), Y_POS, start.getZ()],
                      [end.getX(), Y_POS, end.getZ()]]]
        self.node = LineNodePath()
        self.node.reparentTo(paper)
        self.node.setColor(0.2,0.2,0.2,1)
        self.node.setTransparency(1)
        self.node.setThickness(3)
        self.node.drawLines(self.pointList)
        self.node.setTag('line', '1')

        cs = CollisionSegment(start.getX(),Y_POS,start.getZ(),end.getX(),Y_POS,end.getZ())
        cs.setTangible(True)
        cnode = CollisionNode('line')
        cnode.addSolid(cs)
        cnode.setFromCollideMask(BitMask32(0x10))
        cnode.setIntoCollideMask(BitMask32(0x11))  #so that it doesnt receive collisions
        cnodePath = self.node.attachNewNode(cnode)
        base.cTrav.addCollider(cnodePath, handlerQ)
        
        self.node.create()        
        self.dieInt = LerpColorInterval(self.node, 1, Vec4(0.2,0.2,0.2,0))
        self.killInt = Func(self.kill)
        myS = Sequence(Wait(2.0), self.dieInt, self.killInt)
        #parSeq.append(myS)
        myS.start()

        #parSeq.start()

        
    def kill(self):
        self.node.removeNode()
        del self
    
        
            

## #add button
## b = DirectButton(text = ("OK", "click!", "rolling over", "disabled"),
##                  scale=.05,
##                  command=setText)


class MyWorld(DirectObject):

    def __init__(self):
        self.readConfigFile()
        self.setupPicking()
        self.createWorld()
        self.setupCamera()
        self.setupTasks()
        self.setupHandlers()
        #self.setupLighting()

        # for printing performance (FPS and NumUpdates)
        self.totalFrames = 0
        self.tm = 0
        self.finalTime = 0
        self.runTimer = True
        self.pause = False
        self.pauseTime = 0

        # for mouse
        self.mousebtn = [False,False,False]
        self.mouseClickX = 0
        self.mouseClickY = 0
        self.doPause()
        parSeq.loop()



    def createWorld(self):
        # the background
        self.paper = loader.loadModel(DATA_PATH+"/paper.egg")
        self.paper.reparentTo(render)

        # set up paper collisions
        cp = CollisionPlane(Plane(Vec3(0, -1, 0), Point3(0, 0, 0)))
        cnode = CollisionNode('plane')
        cnode.setTag('plane', '1')
        cnode.addSolid(cp)
        cnode.setIntoCollideMask(BitMask32(0x2))
        cnodePath = self.paper.attachNewNode(cnode)
        #cnodePath.show()
        self.paper.setTag('paper', '1')
        self.heading = self.paper.getH()
        self.pitch = self.paper.getP()
        self.setBounds()

        #self.int1 = LerpHprInterval(self.paper, 2, Vec3(0,10,0), Vec3(0,-10,0), blendType='easeInOut', fluid=1)
        #self.int2 = LerpHprInterval(self.paper, 2, Vec3(0,-10,0), Vec3(0,10,0), blendType='easeInOut', fluid=1)
        #self.int3 = LerpHprInterval(self.paper, 2, Vec3(10,0,0), Vec3(-10,0,0), blendType='easeInOut', fluid=1)
        #self.int4 = LerpHprInterval(self.paper, 2, Vec3(-10,0,0), Vec3(10,0,0), blendType='easeInOut', fluid=1)

        self.int1 = LerpHprInterval(self.paper, 4, Vec3(10,10,0), Vec3(10,-10,0), blendType='easeInOut', fluid=1)
        self.int2 = LerpHprInterval(self.paper, 4, Vec3(-10,-10,0), Vec3(10,10,0), blendType='easeInOut', fluid=1)
        self.int3 = LerpHprInterval(self.paper, 4, Vec3(-10,10,0), Vec3(-10,-10,0), blendType='easeInOut', fluid=1)
        self.int4 = LerpHprInterval(self.paper, 4, Vec3(10,-10,0), Vec3(-10,10,0), blendType='easeInOut', fluid=1)
         
        #self.int1 = LerpHprInterval(self.paper, 2, Vec3(10,10,0), Vec3(-10,-10,0), blendType='easeInOut', fluid=1)
        #self.int2 = LerpHprInterval(self.paper, 2, Vec3(-10,-10,0), Vec3(10,10,0), blendType='easeInOut', fluid=1)
        #seq1 = Sequence(self.int1, self.int2)
        #seq2 = Sequence(self.int3, self.int4)
        #p1 = Parallel(self.int1, self.int3)
        #p2 = Parallel(self.int2, self.int4)
        self.paperSeq = Sequence(self.int1, self.int2, self.int3, self.int4)
        parSeq.append(self.paperSeq)
        #self.paperSeq.loop()

        # text
        self.timeTxt = OnscreenText(text="TIME: 0.0s",
                                    fg=(1,1,0,1),
                                    pos=(0,0.95),
                                    scale = .07,
                                    mayChange=1)

        # button
        self.b = DirectButton(text = ("Start"),scale=.2,command=self.startGame)

        
        # create the bins for the spacial partitioning
        self.bins = Bins(self.paper)

        # the corners (one per letter type)
        self.setupCorners()

        # the magnets (one per corner)
        self.setupMagnets()

        # handler for boids
        self.handlerQ = CollisionHandlerEvent()
        self.handlerQ.addInPattern('into')#'%(line)fh-into-%(boid)ih')
        self.handlerQ.addOutPattern('out')#'%(line)fh-into-%(boid)ih')

        
        # the boids
        self.boids = BoidList(self.paper, self.letters, self.bins, self.magnets, self.corners, self.handlerQ, self)
        self.accept( 'into', self.boids.handleInCollision)
        self.accept( 'out', self.boids.handleOutCollision)
                             
        # lines
        self.lines = []
        #self.doPause()

        
    def startGame(self):
        self.b.destroy()
        self.doPause()
        

    def setupCorners(self):
        self.corners = {}  #keyed by letter
        
        # pick random corners for the letters
        random.seed()
        randCorners = random.sample([0,1,2,3], len(self.letters))
        c = {}
        i = 0
        for letter in self.letters:
            c[letter] = randCorners[i]
            i += 1
            
        for letter, corner in c.iteritems():
            self.corners[letter] = Corner(corner, letter, self.paper)
            
            
    def setupMagnets(self):
        self.magnets = {}

        for letter, cornerObj in self.corners.iteritems():
            self.magnets[letter] = Magnet(letter, cornerObj, self.paper)
        
        
        # figure out the bounds from the extents of the paper
    def setBounds(self):
        global X_MAX, X_MIN, Z_MAX, Z_MIN
        pBounds = self.paper.getTightBounds()
        X_MAX = pBounds[1].getX()
        X_MIN = pBounds[0].getX()
        Z_MAX = pBounds[1].getZ()
        Z_MIN = pBounds[0].getZ()


    def setupLighting(self):
        self.lightAttrib = LightAttrib.makeAllOff() 
        self.spotlight = Spotlight( "spotlight" )
        self.spotlight.setColor( Vec4( .5, .5, .5, 1 ) )
        self.spotlight.setLens( PerspectiveLens() )
        self.spotlight.getLens().setFov( 10, 10 )
        self.spotlight.setAttenuation( Vec3( 1, 0.0, 0.0 ) ) 
        self.spotlight.setExponent( 50.0 )
        camera.attachNewNode( self.spotlight.upcastToLensNode() ) 
        self.lightAttrib = self.lightAttrib.addLight( self.spotlight )
        render.node().setAttrib( self.lightAttrib )


    def setupCamera(self):
        self.camDist = -250
        base.disableMouse()
        base.camera.setPos(0, self.camDist, 0)
        base.camera.lookAt(self.paper.getX(), self.paper.getY(), self.paper.getZ())
        #self.heading = base.camera.getH()
        #self.pitch = base.camera.getP()


    def setupPicking(self):
        self.clickedObj = None  #obj that was picked
        self.paperClickPos = None
        self.paperDeClickPos = None
        
        self.picker = CollisionTraverser()            #Make a traverser
        base.cTrav = self.picker
        self.pq     = CollisionHandlerQueue()         #Make a handler
        self.pickerNode = CollisionNode('mouseRay')
        self.pickerNP = camera.attachNewNode(self.pickerNode)
        self.pickerNode.setFromCollideMask(BitMask32(0x2))
        self.pickerNode.setIntoCollideMask(BitMask32(0x3))
        self.pickerRay = CollisionRay()               #Make our ray
        self.pickerNode.addSolid(self.pickerRay)      #Add it to the collision node
        base.cTrav.addCollider(self.pickerNP, self.pq)
        #self.picker.showCollisions(render)


    def setupTasks(self):
        taskMgr.add(self.boids.updateBoids, "UpdateBoids")
        taskMgr.add(self.printFPS, "FPS")
        taskMgr.add(self.controlCamera, "camera")
        taskMgr.add(self.handleMouseDrag, "drag")


    def setupHandlers(self):
        self.accept("mouse3", self.setMouseBtn, [2, True])
        self.accept("mouse3-up", self.setMouseBtn, [2, False])
        self.accept("mouse2", self.setMouseBtn, [1, True])
        self.accept("mouse2-up", self.setMouseBtn, [1, False])
        self.accept("mouse1", self.setMouseBtn, [0, True])
        self.accept("mouse1-up", self.setMouseBtn, [0, False])
        self.accept("escape", sys.exit, [0])
        self.accept("p", self.doPause)
        self.accept("q", self.bins.showBins)
        self.accept("arrow_up", self.setBoidSpeed, [True])
        self.accept("arrow_down", self.setBoidSpeed, [False])


    def setBoidSpeed(self, speedUp):
        global SPEED_LIMIT
        if speedUp:
            SPEED_LIMIT += 0.5
        else:
            SPEED_LIMIT -= 0.5
            if SPEED_LIMIT < 0.5:
                SPEED_LIMIT = 0.5


    def setMouseBtn(self, btn, value):
        self.mousebtn[btn] = value
        md = base.win.getPointer(0)
        self.mouseClickX = md.getX()
        self.mouseClickY = md.getY()

        if btn == 0:
            mpos = base.mouseWatcherNode.getMouse()
            self.pickerRay.setFromLens(base.camNode, mpos.getX(), mpos.getY())

            base.cTrav.traverse(self.paper)
            if self.pq.getNumEntries() > 0:
                self.pq.sortEntries()
                cEntry = self.pq.getEntry(0)
                nodePath = cEntry.getIntoNodePath().findNetTag("magnet")
                if not nodePath.isEmpty(): # magnet collision
                    if value:  # click
                        self.clickedObj = nodePath
                        self.magnets[nodePath.getTag('magnet')].held = True
                    else:      # declick
                        #self.magnets[self.clickedObj.getTag("magnet")].moveToCorner()
                        self.magnets[nodePath.getTag('magnet')].held = False
                        self.clickedObj = None
                        
                elif not cEntry.getIntoNodePath().findNetTag("paper").isEmpty():
                    pos = cEntry.getSurfacePoint(self.paper)
                    if value:  # click
                        self.paperClickPos = pos
                    elif self.paperClickPos:      # declick
                        if (pos-self.paperClickPos).length() > 5:
                            self.lines.append( Line(self.paperClickPos, pos, self.paper, self.handlerQ) )
                        self.paperClickPos = None
                        self.paperDeClickPos = None
                        

    def getPaperMousePos(self, tag="paper"):
        if not base.mouseWatcherNode.hasMouse():
            return None
        
        mpos = base.mouseWatcherNode.getMouse()
        self.pickerRay.setFromLens(base.camNode, mpos.getX(), mpos.getY())

        base.cTrav.traverse(self.paper)
        if self.pq.getNumEntries() > 0:
            self.pq.sortEntries()
            cEntry = self.pq.getEntry(0)
            nodePath = cEntry.getIntoNodePath().findNetTag(tag)
            if tag == "paper" and not nodePath.isEmpty():
                return cEntry.getSurfacePoint(self.paper)
            elif tag == "magnet" and not nodePath.isEmpty():
                return nodePath
            else:
                return None


    def handleMouseDrag(self, task):
        if not base.mouseWatcherNode.hasMouse():
            return Task.cont
        if self.clickedObj:
            pos = self.getPaperMousePos()
            if pos:
                letter = self.clickedObj.getTag("magnet")
                magnet = self.magnets[letter]
                magnet.setPosition( Vec2D(pos.getX(),pos.getZ()) )

        elif self.paperClickPos:  # to check when we go out of bounds with the line
            pos = self.getPaperMousePos()
            if pos and (pos.getX()>X_MAX or pos.getX()<X_MIN or
                        pos.getZ()<Z_MIN or pos.getZ()>Z_MAX):
                if (pos-self.paperClickPos).length() > 5:
                    self.lines.append( Line(self.paperClickPos, pos, self.paper, self.handlerQ) )
                self.paperClickPos = None
                self.paperDeClickPos = None
                
        #elif self.mousebtn[0]:
            

        return Task.cont


    # loop through the config file and read the letters into a hash
    def readConfigFile(self):
        self.letters = {}  #keyed by letters, value = number of letters
                
        # read the file letter by letter
        total = 0
        f = open(CONFIG_FILE, "r")
        for line in f:
            line = line.strip()
            if line.startswith("#"):  #allow comments
                continue
            for c in line:
                c = c.lower()
                if (not c in string.ascii_lowercase) or \
                    (not c in ACCEPTED_LETTERS):  
                    continue     # skip anything but accepted letters
                if c in self.letters:
                    total += 1
                    self.letters[c] += 1
                else:
                    total += 1
                    self.letters[c] = 1
        f.close()
        #print "\n\nletters to use: ", self.letters, "\n\n"
        #print "total letters: ", total


    def stopTimer(self):
        self.runTimer = False
        self.paperSeq.pause()#finish()
        self.endText = OnscreenText(text="Your time is:"+("%.1f s")%self.finalTime,
                                    fg=(0.2,0.2,0.2,1),
                                    pos=(0,0.2),
                                    scale = .1)
        self.endB = DirectButton(text = ("Quit"),scale=.07,command=sys.exit)
        
    def printFPS(self, task):
        if (task.time - self.tm) < 0.1:
            return Task.cont
        self.tm = task.time

##         # number of frames per second
##         currentFrame = globalClock.getFrameCount()
##         print "FPS: "+ str(currentFrame - self.totalFrames)

##         # number of updates per second
##         global frameCount
##         print "Num Updates: ", frameCount
##         frameCount = 0
##         self.totalFrames = currentFrame

        if self.runTimer:
            if not self.pause:
                self.finalTime = task.time - self.pauseTime
                self.timeTxt.setText("TIME: "+("%.1f")%self.finalTime+" s")
            else:
                self.pauseTime = task.time - self.finalTime
        else:
            return Task.done

        #fpsText.setText("FPS: "+ str(current - totalFrames) )
        return Task.cont



    def doPause(self):
        self.pause = not self.pause
        self.boids.pause()
        if self.pause:
            #self.paperSeq.pause()
            parSeq.pause()
        else:
            parSeq.resume()
            #self.paperSeq.resume()
            

    # controls the camera when the mouse is moved
    def controlCamera(self, task):
        md = base.win.getPointer(0)
        x = md.getX()
        y = md.getY()
        dx = x - self.mouseClickX
        dy = y - self.mouseClickY
        self.mouseClickX = x
        self.mouseClickY = y

        if self.mousebtn[2]:
            self.heading = self.heading + dx*0.5
            self.pitch = self.pitch + dy*0.5
            self.paper.setHpr(self.heading,self.pitch,0)
        elif self.mousebtn[1]:
            self.camDist += dy*0.4
            if self.camDist > -50: self.camDist = -50
            base.camera.setPos(0, self.camDist, 0)
        return Task.cont



# do the JIT compiling to speed things up a bit
psyco.full()

## wp = WindowProperties()
## wp.setFullscreen(1)
## base.openMainWindow(props = wp)

# create our world
MyWorld()

# run the main loop
run()
